import sys
import os
import pytest
import pandas as pd 
import arcus.ml.neuralnetworks.keras as nnk
import logging

logging.basicConfig(level=logging.DEBUG)
mylogger = logging.getLogger()

def check_gpu():
    from tensorflow.python.client import device_lib

    local_device_protos = device_lib.list_local_devices()
    return len([x.name for x in local_device_protos if x.device_type == 'GPU']) > 0

def test_gpu():
    if not check_gpu():
        import pytest
        pytest.skip('Test requires GPU')

    # print dataframe. 
    assert nnk.enable_gpu() == True




'''
The timeops module provides interesting standard functionality for timeseries datasets
'''

import pandas as pd
import math
import numpy as np
import datetime

def set_timeseries(df: pd.DataFrame, time_column:str) -> pd.DataFrame:
    '''
    Transforms the dataframe to a timeseries enabled dataframe
    by applying a DateTimeIndex to the given column

    Args:
        df (pd.DataFrame): The DataFrame that should be time-indexed
    
    Returns:
        pd.DataFrame: The DataFrame, including the DatetimeIndex
    '''
    df.index = pd.DatetimeIndex(df[time_column])
    pd.to_datetime(df[time_column], errors='coerce')
    return df

def add_time_reference(sorted_df: pd.DataFrame, n: int, reference_column: str, new_column: str) -> pd.DataFrame:
    '''
    This method will add a reference column to the DataFrame that points to n items before
    Args:
        sorted_df (pd.DataFrame): a DataFrame that is sorted by the time
        n (int): the number of records to look back for the referencing columns
        new_column (str): the name of the new column
    Returns:
        (pd.DataFrame): the dataframe with the new column created
    '''
    sorted_df[new_column] = sorted_df[reference_column].shift(n)

def time_slice(df: pd.DataFrame, time_column:str, start: datetime.datetime = None, end: datetime.datetime = None):
    '''
    This method takes a time series DataFrame and only returns the time slice, based on the start & end date
    Args:
        df (pd.DataFrame): The indexed Data Frame , which should be a DatetimeIndex
        start (datetime): The start time of the time slice.  When skipped, the time slice begins at the beginning of the index
        end (datetime): The end time of the time slice.  When skipped, the time slice end at the end of the index
    Returns:
        (pd.DataFrame): the dataframe only containing the records inside the time slice
    '''
    if(df.index is None or type(df.index) != pd.DatetimeIndex):
        df = set_timeseries(df, time_column)
    df = df.sort_index()
    return df.loc[start:end]

def get_windows(sorted_df: pd.DataFrame, window_size: int, window_stride: int = 1, group_column: str = None, zero_padding: bool = False, remove_group_column: bool = False, target_column: str = None) -> np.array:
    if group_column is None:
        return __get_windows_from_group(sorted_df, window_size, window_stride, zero_padding, target_column)

    else:
        windows = None
        targets = None
        
        # create unique list of groups
        _groups = sorted_df[group_column].unique()

        for key in _groups:
            _group_df = sorted_df[:][sorted_df[group_column] == key]
            if(remove_group_column):
                _group_df.drop(group_column, axis=1, inplace=True)
            _current_windows, _current_targets = __get_windows_from_group(
                _group_df, window_size, window_stride, zero_padding, target_column)
            if windows is None:
                windows = _current_windows
            else:
                windows = np.concatenate((windows, _current_windows))

            if target_column is not None:
                if targets is None:
                    targets = _current_targets
                else:
                    targets = np.concatenate((targets, _current_targets))

        return windows, targets

def __get_windows_from_group(sorted_df: pd.DataFrame, window_size: int, 
                             window_stride: int = 1, zero_padding: bool = False,
                             target_column:str = None) -> np.array:
    # zero based row to take the leading window from - this row will be last row in the window
    _start_row_idx = 0 if zero_padding else window_size - 1
    # will contain all windows
    windows = list()
    targets = list()
    
    # range will be from start_row to row_count
    for __current_row_idx in range(_start_row_idx, len(sorted_df)):
        __slice_begin_idx = (__current_row_idx - window_size +
                             1) if __current_row_idx >= window_size else 0
        window_df = sorted_df.copy()
        time_slice = window_df.iloc[__slice_begin_idx:__current_row_idx + 1, :]
        time_array = np.array(time_slice.values)
        if(zero_padding):
            _rows_to_pad = window_size - __current_row_idx - 1
            if(_rows_to_pad > 0):
                padding_matrix = np.zeros(
                    (_rows_to_pad, len(sorted_df.columns)))
                time_array = np.concatenate((padding_matrix, time_array))
        if(target_column is not None):
            _target_colidx = sorted_df.columns.get_loc(target_column)
            time_array, target_array = __pop_from_array(time_array, _target_colidx)
            targets.append(target_array)

        windows.append(time_array)
    print(len(targets))
    return np.array(windows), np.array(targets) if target_column is not None else None

def __pop_from_array(my_array,pc):
    i = pc
    pop = my_array[:,i]
    new_array = np.hstack((my_array[:,:i],my_array[:,i+1:]))
    return new_array, pop

def combine_time_ranges(*args: pd.DataFrame):
    _result_df = pd.concat(args)
    _result_df = _result_df.drop_duplicates()
    return _result_df


import sys
import os
import pytest
import pandas as pd 
import arcus.ml.timeseries.timeops as tops
import arcus.ml.dataframes as adf
import logging
from unittest.mock import patch 
from datetime import datetime
import numpy as np 

logging.basicConfig(level=logging.DEBUG)
mylogger = logging.getLogger()

time_series_df = pd.read_csv('tests/resources/datasets/sunspots.csv')

def test_time_slice_default_str():
    df = time_series_df.copy()
    slice_df = tops.time_slice(df, 'Date')
    assert slice_df.shape == (30, 3)
    #df = tops.time_slice(df, '')

def test_time_slice_start_str():
    df = time_series_df.copy()
    slice_df = tops.time_slice(df, 'Date', start='01/02/2001')
    assert slice_df.shape == (30, 3)
    slice_df = tops.time_slice(df, 'Date', start='01/01/2018')
    assert slice_df.shape == (5, 3)

def test_time_slice_end_str():
    df = time_series_df.copy()
    slice_df = tops.time_slice(df, 'Date', end='01/02/2020')
    assert slice_df.shape == (30, 3)
    slice_df = tops.time_slice(df, 'Date', end='01/01/2017')
    assert slice_df.shape == (13, 3)

def test_time_slice_both_str():
    df = time_series_df.copy()
    slice_df = tops.time_slice(df, 'Date', start = '01/01/2001', end='01/02/2020')
    assert slice_df.shape == (30, 3)
    slice_df = tops.time_slice(df, 'Date', start='01/01/2016', end='01/01/2017')
    assert slice_df.shape == (12, 3)


def test_time_slice_start_date():
    df = time_series_df.copy()
    slice_df = tops.time_slice(df, 'Date', start=datetime.strptime('01/02/2001', '%d/%m/%Y'))
    assert slice_df.shape == (30, 3)
    slice_df = tops.time_slice(df, 'Date', start=datetime.strptime('01/01/2018', '%d/%m/%Y'))
    assert slice_df.shape == (5, 3)

def test_time_slice_end_date():
    df = time_series_df.copy()
    slice_df = tops.time_slice(df, 'Date', end=datetime.strptime('01/02/2020', '%d/%m/%Y'))
    assert slice_df.shape == (30, 3)
    slice_df = tops.time_slice(df, 'Date', end=datetime.strptime('01/01/2017', '%d/%m/%Y'))
    assert slice_df.shape == (13, 3)

def test_time_slice_both_date():
    df = time_series_df.copy()
    slice_df = tops.time_slice(df, 'Date', start=datetime.strptime('01/01/2001', '%d/%m/%Y'), end=datetime.strptime('01/02/2020', '%d/%m/%Y'))
    assert slice_df.shape == (30, 3)
    slice_df = tops.time_slice(df, 'Date', start=datetime.strptime('01/01/2016', '%d/%m/%Y'), end=datetime.strptime('01/01/2017', '%d/%m/%Y'))
    assert slice_df.shape == (12, 3)

def test_add_timerange():
    range01 = pd.date_range('2018-03-01', periods=6, freq='1M')
    df01 = pd.DataFrame({'A': np.ones(6), 'Timestamp': range01})
    range02 = pd.date_range('2018-01-01', periods=6, freq='1M')
    df02 = pd.DataFrame({'A': np.ones(6), 'Timestamp': range02})
    range03 = pd.date_range('2018-05-01', periods=6, freq='1M')
    df03 = pd.DataFrame({'A': np.ones(6), 'Timestamp': range03})

    full_range = tops.combine_time_ranges(df01, df02, df03)
    assert len(full_range) == 10

def test_add_empty_df_timerange():
    df00 = None
    range01 = pd.date_range('2018-03-01', periods=6, freq='1M')
    df01 = pd.DataFrame({'A': np.ones(6), 'Timestamp': range01})

    full_range = tops.combine_time_ranges(df00, df01)
    assert len(full_range) == 6

def test_time_windows_default():
    df = pd.read_csv('tests/resources/datasets/engine-sensors.csv')
    assert len(df)==20  
    assert df.shape == (20, 27)  
    win_size = 6
    col_size = len(df.columns)
    windows, _ = tops.get_windows(df, win_size)
    assert windows.shape == (15, win_size, col_size)

def test_time_windows_targets():
    df = pd.read_csv('tests/resources/datasets/engine-sensors.csv')
    assert len(df)==20  
    assert df.shape == (20, 27)  
    win_size = 6
    col_size = len(df.columns)
    windows, targets = tops.get_windows(df, win_size, target_column='ttf')
    assert windows.shape == (15, win_size, col_size - 1)
    assert targets.shape == (15, win_size)

def test_time_windows_group_targets():
    df = pd.read_csv('tests/resources/datasets/engine-sensors.csv')
    assert len(df)==20  
    assert df.shape == (20, 27)  
    win_size = 6
    col_size = len(df.columns)
    windows, targets = tops.get_windows(df, win_size, group_column='engine_id', remove_group_column=True, target_column='ttf')
    assert windows.shape == (10, win_size, col_size - 2)
    assert targets.shape == (10, win_size)

def test_time_windows_padding():
    df = pd.read_csv('tests/resources/datasets/engine-sensors.csv')
    assert len(df)==20  
    assert df.shape == (20, 27)  
    win_size = 6
    col_size = len(df.columns)
    windows, _ = tops.get_windows(df, win_size, zero_padding=True)
    assert windows.shape == (20, win_size, col_size)

def test_time_windows_group():
    df = pd.read_csv('tests/resources/datasets/engine-sensors.csv')
    win_size = 6
    col_size = len(df.columns)
    windows, _ = tops.get_windows(df, win_size, group_column='engine_id')
    assert windows.shape == (10, win_size, col_size)

def test_time_windows_group_padding():
    df = pd.read_csv('tests/resources/datasets/engine-sensors.csv')
    win_size = 6
    col_size = len(df.columns)
    windows, _ = tops.get_windows(df, win_size, group_column='engine_id', zero_padding=True)
    assert windows.shape == (20, win_size, col_size)



import sys
import os
import pytest
import pandas as pd 
import logging
import numpy as np
from unittest.mock import patch 
from sklearn.model_selection import train_test_split
from sklearn import linear_model
from arcus.ml.evaluation import classification as clev

def test_binary_classifier_evaluation():
    df = pd.read_csv('tests/resources/datasets/student-admission.csv')
    y = df.Admission.values
    X = np.asarray(df.drop(['Admission'],axis=1))
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.30, random_state=0)

    # trainen van de logistic regression classifier
    logreg = linear_model.LogisticRegression(C=1e5,solver='liblinear')
    logreg.fit(X_train, y_train)
    clev.evaluate_model(logreg, X_test, y_test)

@patch("matplotlib.pyplot.show")
def test_binary_classifier_evaluation_roc(mock_show):
    df = pd.read_csv('tests/resources/datasets/student-admission.csv')
    y = df.Admission.values
    X = np.asarray(df.drop(['Admission'],axis=1))
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.30, random_state=0)

    # trainen van de logistic regression classifier
    logreg = linear_model.LogisticRegression(C=1e5,solver='liblinear')
    logreg.fit(X_train, y_train)
    clev.evaluate_model(logreg, X_test, y_test, True)
    

def test_multiclass_classifier_evaluation_roc():
    df = pd.read_csv('tests/resources/datasets/wine-makers.csv')
    y = df.Cultivar.values
    X = df.drop(['Cultivar'],axis=1)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.30, random_state=0)

        # trainen van de logistic regression classifier
    logreg = linear_model.LogisticRegression(C=1,solver='lbfgs')
    logreg.fit(X_train, y_train)
    # We are trying to apply the Roc curve for a multi class classifier
    # This should raise an exception
    with pytest.raises(AttributeError) as attribute_err:
        clev.evaluate_model(logreg, X_test, y_test, True)
    assert 'binary classifier' in str(attribute_err.value)

def test_multiclass_classifier_evaluation_noroc():
    df = pd.read_csv('tests/resources/datasets/wine-makers.csv')
    y = df.Cultivar.values
    X = df.drop(['Cultivar'],axis=1)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.30, random_state=0)

        # trainen van de logistic regression classifier
    logreg = linear_model.LogisticRegression(C=1,solver='lbfgs')
    logreg.fit(X_train, y_train)

    clev.evaluate_model(logreg, X_test, y_test, False)
